﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using OpenTK.Graphics.OpenGL;
using OpenTK.Math;
using Matrix4=OpenTK.Matrix4;
using Vector3=OpenTK.Vector3;
using Vector4=OpenTK.Vector4;

namespace TKTest {
    [Flags]
    public enum KButtons {
        ZERO=0,
        M_LEFT=1,
        M_RIGHT=2,
        MOUSE=M_LEFT|M_RIGHT,
        K_SHIFT=4,
        K_CTRL=8,
        K_ALT=16
    }


    class Scene3D {
        int m_w,m_h;
        Camera3D m_cam;
        CGame m_game;
        int m_id;

        double m_angle=Math.PI/4;

        uint[] m_vbo=new uint[1];

        public Scene3D(CGame _game,int _id) {
            m_game=_game;
            m_id=_id;
            m_cam=new Camera3D(_id==0 ? 20 : 4);
        }

        public void InitView(int _w,int _h) {
            GL.ClearColor(Color.Black);
            GL.Enable(EnableCap.DepthTest);
            GL.EnableClientState(EnableCap.CullFace);
            GL.Enable(EnableCap.Blend);
            GL.BlendFunc(BlendingFactorSrc.SrcAlpha,BlendingFactorDest.OneMinusSrcAlpha);
            GL.EnableClientState(EnableCap.AlphaTest);
            GL.DepthFunc(DepthFunction.Less);
//            GL.AlphaFunc(AlphaFunction.Greater,0.2f);
            GL.ClearDepth(1);
            m_w=_w; 
            m_h=_h;
            SetupViewport();

            GL.EnableClientState(EnableCap.ColorArray);
            GL.EnableClientState(EnableCap.VertexArray);
            GL.EnableClientState(EnableCap.NormalArray);
            SetLight();

            GL.GenBuffers(1,m_vbo);
        }

        public void SetSize(int _w,int _h) {
            m_w=_w;
            m_h=_h;
            SetupViewport();
        }


        bool m_mouseLeftPressed=false;
        bool m_mouseRightPressed=false;
        int m_lastMouseX,m_lastMouseY;
        KButtons m_lastMouseKBtn;
        int m_mouseWay;

        const int MOUSE_WAY_LIMIT=6;

        internal void MouseDown(bool _qleft,int _x,int _y,KButtons _kbd) {
            if(_qleft) m_mouseLeftPressed=true;
            else m_mouseRightPressed=true;
            if(m_mouseLeftPressed) _kbd|=KButtons.M_LEFT;
            if(m_mouseRightPressed) _kbd|=KButtons.M_RIGHT;

            m_lastMouseX=_x;
            m_lastMouseY=_y;
            m_lastMouseKBtn=_kbd;
            m_mouseWay=0;
        }
        internal void MouseLeave() {
            m_mouseLeftPressed=m_mouseRightPressed=false;
        }

        internal void MouseMove(int _x,int _y,KButtons _kbd) {
            if(!m_mouseRightPressed && !m_mouseLeftPressed) return;
            if(m_mouseLeftPressed) _kbd|=KButtons.M_LEFT;
            if(m_mouseRightPressed) _kbd|=KButtons.M_RIGHT;
            int dx=_x-m_lastMouseX,dy=_y-m_lastMouseY;
            m_mouseWay+=Math.Abs(dx)+Math.Abs(dy);
            ProcessMouseMove(_kbd,_x,_y,dx,dy);
            m_lastMouseY=_y; m_lastMouseX=_x;
        }

        internal void MouseUp(bool _qleft,int _x,int _y,KButtons _kbd) {
            if(_qleft && !m_mouseLeftPressed) return;
            if(!_qleft && !m_mouseRightPressed) return;
            if(m_mouseLeftPressed) _kbd|=KButtons.M_LEFT;
            if(m_mouseRightPressed) _kbd|=KButtons.M_RIGHT;
            if(m_mouseWay<MOUSE_WAY_LIMIT) {
                ProcessMouseClick(_kbd,_x,_y);
            }
            m_mouseLeftPressed=m_mouseRightPressed=false;
        }





        void ProcessMouseMove(KButtons _btns,int _x0,int _y0,int _dx,int _dy) {
            double dx,dy;
            if(_btns==KButtons.M_RIGHT) {  // turn head left/right
                dx=(double)_dx*m_angle/m_h;
                dy=(double)_dy*m_angle/m_h;
            } else {
                dx=(double)_dx*3/m_h;
                dy=(double)_dy*3/m_h;
            }
            ProcessMouseMoveF(_btns,dx,dy);
        }

        public void ProcessMouseClick(KButtons _btns,int _x,int _y) {
            if(m_id==1) {
                m_game.SwapFig();
                return;
            }


            // 0,0,m_w,m_h
            float x=(float)((2.0*_x)/m_w-1);
            float y=(float)(1-(2.0*_y)/m_h);

            Matrix4 mv,mp;
            GL.GetFloat(GetPName.ModelviewMatrix,out mv);
            GL.GetFloat(GetPName.ProjectionMatrix,out mp);
            Matrix4 inv=(mv*mp);
            inv.Invert();
            Vector4 v1=Vector4.Transform(new Vector4(x,y,0,1),inv);

            Vector3 orig=m_cam.Pos;
            Vector3 dir=new Vector3(v1.X/v1.W-orig.X,v1.Y/v1.W-orig.Y,v1.Z/v1.W-orig.Z);

            m_game.ProcessClick(orig,dir);
        }


        internal void ProcessMouseMoveF(KButtons _btns,double hx,double hy) {

            switch(_btns) {
                case KButtons.M_LEFT: // (x,y)-w rotate around object
                    m_cam.Shift(1);
                    m_cam.Rotate(-hx,1,0);
                    m_cam.Rotate(-hy,2,0);
                    m_cam.Shift(-1);
                    break;
                case KButtons.M_LEFT|KButtons.K_CTRL: // z-w rotate around object and zoom
                    m_cam.Shift(1);
                    m_cam.Scale(hy);
                    m_cam.Shift(-1);
                    break;
                case KButtons.M_RIGHT: // (x,y)-w rotate around camera
                    m_cam.Rotate(hx,1,0);
                    m_cam.Rotate(hy,2,0);
                    break;
            }
        }

        public void RestoreView() {
            m_cam.Restore();
        }

        public void Render() {
            GL.MatrixMode(MatrixMode.Modelview);
            GL.LoadMatrix(ref Matrix4.Identity);
            GL.Light(LightName.Light0,LightParameter.Position,new float[] { 2f,3f,-10f });
            Matrix4 lookat = m_cam.GetMatrix();
            GL.LoadMatrix(ref lookat);
            DrawMesh();
            m_game.SetMatrix(m_id,lookat);
        }


        void SetLight() {
            // Enable Light 0 and set its parameters.
            GL.Light(LightName.Light0,LightParameter.Position,new float[] { 2f,3f,-10f });
            GL.Light(LightName.Light0,LightParameter.Ambient,new float[] { 0.3f,0.3f,0.3f,0.2f });
            GL.Light(LightName.Light0,LightParameter.Diffuse,new float[] { 1.0f,1.0f,1.0f,0.2f });
            GL.Light(LightName.Light0,LightParameter.Specular,new float[] { 1.0f,1.0f,1.0f,0.2f });
//            GL.Light(LightName.Light0,LightParameter.SpotExponent,new float[] { 1.0f,1.0f,1.0f,0.2f });
            GL.LightModel(LightModelParameter.LightModelAmbient,new float[] { 0.2f,0.2f,0.2f,0.2f });
            GL.LightModel(LightModelParameter.LightModelTwoSide,0);
            GL.LightModel(LightModelParameter.LightModelLocalViewer,1);
            GL.LightModel(LightModelParameter.LightModelColorControl,(float)(int)LightModelColorControl.SeparateSpecularColor);
            GL.Enable(EnableCap.Lighting);
            GL.Enable(EnableCap.Light0);
            GL.Enable(EnableCap.ColorMaterial);
            GL.ShadeModel(ShadingModel.Smooth);
            // Use GL.Material to set your object's material parameters.
            GL.ColorMaterial(MaterialFace.Front,ColorMaterialParameter.AmbientAndDiffuse);
            
        }
        private void SetupViewport() {
            GL.Viewport(0,0,m_w,m_h);
            float aspect_ratio = (float)m_w/m_h;
            Matrix4 perpective = Matrix4.CreatePerspectiveFieldOfView((float)m_angle,aspect_ratio,0.1f,100);
            GL.MatrixMode(MatrixMode.Projection);
            GL.LoadMatrix(ref perpective);
        }

        private void DrawMesh() {
            if(m_game==null) return;
            MyVertex[] trgs;
            int ntrg,nseg;
            trgs=m_game.GetBuffer(m_cam.Pos,m_cam.Dir,m_id,out ntrg,out nseg);
            int lv=MyVertex.SizeInBytes;

            GL.BindBuffer(BufferTarget.ArrayBuffer,m_vbo[0]);
            GL.ColorPointer(4,ColorPointerType.UnsignedByte,MyVertex.SizeInBytes,(IntPtr)0);
            GL.NormalPointer(NormalPointerType.Float,MyVertex.SizeInBytes,(IntPtr)4);
            GL.VertexPointer(3,VertexPointerType.Float,MyVertex.SizeInBytes,(IntPtr)16);

            int NP=3*ntrg+2*nseg;
            GL.BufferData(BufferTarget.ArrayBuffer,(IntPtr)(lv*NP),IntPtr.Zero,BufferUsageHint.StreamDraw);
            GL.BufferData(BufferTarget.ArrayBuffer,(IntPtr)(lv*NP),trgs,BufferUsageHint.StreamDraw);
            if(ntrg!=0) GL.DrawArrays(BeginMode.Triangles,0,3*ntrg);
            if(nseg!=0) GL.DrawArrays(BeginMode.Lines,3*ntrg,2*nseg);

            GL.Finish();
        }

    }

    
    public struct MyVertex {
        public byte R,G,B,A;
        public Vector3 Normal;
        public Vector3 Position;

        public static int SizeInBytes = 28;

        public MyVertex(ref V3 v,ref V3 n,Color col) {
            Position=new Vector3((float)v.X,(float)v.Y,(float)v.Z);
            Normal=new Vector3((float)n.X,(float)n.Y,(float)n.Z);
            R=(byte)col.R;
            G=(byte)col.G;
            B=(byte)col.B;
            A=(byte)col.A;
        }
    }

    class Camera3D {
        Vector3 m_pos;
        Vector3[] m_basis;
        double m_FLen,m_FLen0;

        Matrix4 m_view;

        public Camera3D(double _flen) {
            m_FLen0=_flen;
            Restore();
        }

        public Vector3 Pos { get { return m_pos; } }
        public Vector3 Dir { get { return m_basis[0]; } }
       
        public Matrix4 GetMatrix() {
            return m_view;
        }

        public void Shift(double _l) {
            m_pos+=m_basis[0]*(float)(_l*m_FLen);
            SetView();
        }
        public void Rotate(double _d,int _x,int _y) {
            _d=Math.Min(0.5,Math.Max(-0.5,_d));
            double _v=Math.Sqrt(1-_d*_d);
            Vector3 v0=m_basis[_x]*(float)_v-m_basis[_y]*(float)_d;
            m_basis[_y]=m_basis[_y]*(float)_v+m_basis[_x]*(float)_d;
            m_basis[_x]=v0;
            SetView();
        }
        public void Scale(double _cf) {
            _cf=Math.Min(1,Math.Max(-0.5,_cf));
            m_FLen*=(1+_cf); 
        }

        internal void Restore() {
            m_pos=new Vector3(20,15,18);
            Vector3 view=new Vector3(-0.7f,-0.8f,-0.5f);
            view.Normalize();
            m_pos=view;
            m_pos.Mult((float)(-m_FLen0));
            m_FLen=m_FLen0;
            Vector3 up=new Vector3(view.Y,-view.X,0);
            up.Normalize();
            Vector3 right=Vector3.Cross(up,view);
            m_basis=new Vector3[]{view,right,up};
            SetView();
        }
        void SetView(){
            Vector3 tg=m_pos+m_basis[0]*(float)m_FLen;
            m_view=Matrix4.LookAt(m_pos,tg,m_basis[2]);
        }
    }
    public struct V3{
        double m_x,m_y,m_z;
        public V3(double x,double y,double z){
            m_x=x; m_y=y; m_z=z;
        }
        static double pyth(double x,double y,double z){
            return Math.Sqrt(x*x+y*y+z*z);
        }
        public double Length(){
            return pyth(m_x,m_y,m_z);
        }
        public double Dist(ref V3 v){
            return pyth(m_x-v.m_x,m_y-v.m_y,m_z-v.m_z);
        }
        public void Norm(){
            double l=Length();
            if(l!=0) NMul(1/l);
        }
        public void NMul(double cf){
            m_x*=cf; m_y*=cf; m_z*=cf;
        }
        public void AddMul(ref V3 v,double cf){
            m_x+=v.m_x*cf;
            m_y+=v.m_y*cf;
            m_z+=v.m_z*cf;
        }
        public double CProd(ref V3 v) {
            return m_x*v.m_x+m_y*v.m_y+m_z*v.m_z;
        }
        public double CProd(ref Vector3 v) {
            return m_x*v.X+m_y*v.Y+m_z*v.Z;
        }
        public V3 VProd(ref V3 v) {
            return new V3(
                m_y*v.m_z-m_z*v.m_y,
                m_z*v.m_x-m_x*v.m_z,
                m_x*v.m_y-m_y*v.m_x);
        }
        public void Ort(ref V3 v) {
            AddMul(ref v,-CProd(ref v));
            Norm();
        }
        public static V3 GetNormal(ref V3 a,ref V3 c,ref V3 b) {
            double bx=b.X-a.X,by=b.Y-a.Y,bz=b.Z-a.Z;
            double cx=c.X-a.X,cy=c.Y-a.Y,cz=c.Z-a.Z;
            V3 res=new V3(by*cz-bz*cy,bz*cx-bx*cz,bx*cy-by*cx);
            res.Norm();
            return res;
        }
        public float Xf { get { return (float)m_x; } }
        public float Yf { get { return (float)m_y; } }
        public float Zf { get { return (float)m_z; } }
        public double X { get { return m_x; } }
        public double Y { get { return m_y; } }
        public double Z { get { return m_z; } }

    }
}
